package net.locusworks.common.crypto;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.security.Key;
import java.security.PrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.Arrays;
import java.util.Base64;
import java.util.List;

import net.locusworks.common.Charsets;
import net.locusworks.common.io.IOUtils;
import net.locusworks.common.utils.DataOutputStreamHelper;
import net.locusworks.common.utils.Utils;

import static net.locusworks.common.utils.Utils.handleExceptionWrapper;
import static net.locusworks.common.utils.Splitter.fixedLengthSplit;
import static java.lang.String.join;

public class KeyFile implements AutoCloseable {
  
  public enum EncryptionType {
    RSA,
    AES,
    SSH
  }
  
  private Key key;
  private String description;
  private Writer writer;
  private EncryptionType encryptionType;
  
  public KeyFile(Key key) {
    this(key, null);
  }
  
  public KeyFile(Key key, String description) {
    this(key, description, EncryptionType.valueOf(key.getAlgorithm().toUpperCase()));
  }
  
  public KeyFile(Key key, String description, EncryptionType encryptionType) {
    this.key = key;
    this.description = description;
    this.encryptionType = encryptionType;
  }
  
  private KeyFile() {}
  
  private void loadFromFile(String fileName) {
    if (Utils.isEmptyString(fileName)) return;
    
    this.key = null;
    try {
      File keyFile = new File(fileName);
      if (!keyFile.exists()) {
        throw new IllegalArgumentException(String.format("Unable to find file with name %s. Please check path", fileName));
      }
      
      String contentStr = IOUtils.toString(new FileInputStream(keyFile), Charsets.UTF_8);
      
      boolean rsaFormat = !contentStr.startsWith("ssh-rsa") && !contentStr.startsWith("aes-seed");
      if (rsaFormat) {
        contentStr = contentStr.replace("-----.*", "");
      }
      
      contentStr = contentStr.replace("\\r?\\n", "");
      
      byte[] content = rsaFormat ? Base64.getDecoder().decode(contentStr): contentStr.getBytes(Charsets.UTF_8);
      
      EncryptionKeyFactory kf = EncryptionKeyFactory.getInstance("RSA");
      
      List<KeySpecHelper> keySpecs = Utils.toList(
          new KeySpecHelper(new AESKeySpec(content), true, EncryptionType.AES),
          new KeySpecHelper(new SSHEncodedKeySpec(content), false, EncryptionType.SSH),
          new KeySpecHelper(new PKCS8EncodedKeySpec(content), true, EncryptionType.RSA),
          new KeySpecHelper(new X509EncodedKeySpec(content), false, EncryptionType.RSA)
          );
      
      for (KeySpecHelper ksh : keySpecs) {
        try {
          this.key = ksh.isPrivate() ? kf.generatePrivateKey(ksh.getKeySpec()) : kf.generatePublicKey(ksh.getKeySpec());
          this.encryptionType = ksh.getEncryptionType();
          break;
        } catch (NullPointerException | InvalidKeySpecException ikse) { continue; }
      }
      
      throw new InvalidKeySpecException(String.format("Unable to determine if file %s is a private or public key. Not type of PKCS8, X509, SSH or AES spec", fileName));
    } catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }
  
  public void write(String fileName) {
    try {
      String data;
      switch(this.encryptionType) {
      case AES:
        data = String.format("%s %s", this.key.getFormat(), Base64.getEncoder().encodeToString(this.key.getEncoded()));
        IOUtils.writeStringToFile(fileName, data);
        break;
      case SSH:
        try (DataOutputStreamHelper dosh = new DataOutputStreamHelper()) {
          getSSHPubKeyBytes(this.key)
            .forEach(handleExceptionWrapper(item ->{
              dosh.writeInt(item.length);
              dosh.write(item);
            }));
          data = String.format("ssh-rsa", dosh.base64Encoded(), this.description);
          IOUtils.writeStringToFile(fileName, data);
        }
        break;
        default:
          writePem(fileName);
      }
    } catch (Exception ex) {
      throw new IllegalArgumentException(ex);
    }
  }
  
  public void setDescription(String description) {
    this.description = description;
  }
  
  public String getDescription() {
    if (Utils.isEmptyString(this.description)) {
      return this.key instanceof PrivateKey ? "PRIVATE KEY" : "PUBLIC KEY"; 
    }
    return this.description;
  }
  
  public Key getKey() {
    return this.key;
  }
  
  @Override
  public void close() {
    try {
      if (this.writer != null) {
        this.writer.flush();
        this.writer.close();
        this.writer = null;
      }
    } catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }
  
  public static KeyFile read(String fileName) {
    try (KeyFile kf = new KeyFile()) {
      kf.loadFromFile(fileName);
      return kf;
    }
  }
  
  private void writePem(String fileName) {
    try {
      String desc = getDescription();
      this.writer = new OutputStreamWriter(new FileOutputStream(fileName), Charsets.UTF_8);
      this.writer.write(String.format("-----BEGIN RSA %s-----", desc));
      
      String encoded = Base64.getEncoder().encodeToString(this.key.getEncoded());
      String out = join("\n", fixedLengthSplit(60).split(encoded));
      
      this.writer.write(out);
      this.writer.write(String.format("-----END RSA %s-----", desc));
    } catch (IOException ex) {
      throw new RuntimeException(ex);
    }
  }
  
  private List<byte[]> getSSHPubKeyBytes(Key key) {
    RSAPublicKey rpk = (RSAPublicKey)key;
    return Arrays.asList("ssh-rsa".getBytes(Charsets.UTF_8),
        rpk.getPublicExponent().toByteArray(),
        rpk.getModulus().toByteArray()
        );
  }
  
  private static class KeySpecHelper {
    private KeySpec keySpec;
    private boolean isPrivate;
    private EncryptionType encryptionType;
    
    public KeySpecHelper(KeySpec keySpec, boolean isPrivate, EncryptionType encryptionType) {
      super();
      this.keySpec = keySpec;
      this.isPrivate = isPrivate;
      this.encryptionType = encryptionType;
    }

    public synchronized final KeySpec getKeySpec() {
      return keySpec;
    }

    public synchronized final boolean isPrivate() {
      return isPrivate;
    }

    public synchronized final EncryptionType getEncryptionType() {
      return encryptionType;
    }
  }

}
